---
id: local-data-updates
title: Local Data Updates
slug: /guided-tour/updating-data/local-data-updates/
---

import DocsRating from '../../../src/core/DocsRating';
import {OssOnly, FbInternalOnly} from 'internaldocs-fb-helpers';

## Local data updates

There are a couple of APIs that Relay provides in order to make purely local updates to the Relay store (i.e. updates not tied to a server operation).

Note that local data updates can be made both on [client-only data](../client-only-data/), or on regular data that was fetched from the server via an operation.

### commitLocalUpdate

To make updates using an [`updater`](../graphql-mutations/#updater-functions) function, you can use the `commitLocalUpdate` API:

```js
import type {Environment} from 'react-relay';

const {commitLocalUpdate, graphql} = require('react-relay');

function commitCommentCreateLocally(
  environment: Environment,
  feedbackID: string,
) {
  return commitLocalUpdate(environment, store => {
    const feedbackRecord = store.get(feedbackID);
    const connectionRecord = ConnectionHandler.getConnection(
      userRecord,
      'CommentsComponent_comments_connection',
    );

    // Create a new local Comment from scratch
    const id = `client:new_comment:${randomID()}`;
    const newCommentRecord = store.create(id, 'Comment');

    // ... update new comment with content

    // Create new edge from scratch
    const newEdge = ConnectionHandler.createEdge(
      store,
      connectionRecord,
      newCommentRecord,
      'CommentEdge' /* GraphQl Type for edge */,
    );

    // Add edge to the end of the connection
    ConnectionHandler.insertEdgeAfter(connectionRecord, newEdge);
  });
}

module.exports = {commit: commitCommentCreateLocally};
```

* `commitLocalUpdate` update simply takes an environment and an updater function.
    * `updater` takes a *`store`* argument, which is an instance of a [`RecordSourceSelectorProxy`](../../../api-reference/store/);  this interface allows you to *imperatively* write and read data directly to and from the Relay store. This means that you have full control over how to update the store: you can *create entirely new records*, or *update or delete existing ones*.
* In our specific example, we're adding a new comment to our local store when. Specifically, we're adding a new item to a connection; for more details on the specifics of how that works, check out our [Adding and Removing Items from a Connection](../../list-data/adding-and-removing-items/) section.
* Note that any local data updates will automatically cause components subscribed to the data to be notified of the change and re-render.

### commitPayload

`commitPayload` takes an `OperationDescriptor` and the payload for the query, and writes it to the Relay Store. The payload will be resolved like a normal server response for a query, and will also resolve Data Driven Dependencies that are passed as `JSResource`, `requireDefer`, etc.

```js
import type {FooQueryRawResponse} from 'FooQuery.graphql'

const {createOperationDescriptor} = require('relay-runtime');

const operationDescriptor = createOperationDescriptor(FooQuery, {
  id: 'an-id',
  otherVariable: 'value',
});

const payload: FooQueryRawResponse = {...};

environment.commitPayload(operation, payload);
```

* An `OperationDescriptor` can be created by `createOperationDescriptor`; it takes the query and the query variables.
* The payload can be typed using the Flow type generated by adding  `@raw_response_type` to the query.
* Note that any local data updates will automatically cause components subscribed to the data to be notified of the change and re-render.

<FbInternalOnly>

### Local editing / creation preview flow


Relay provides several primitives for achieving the preview flow, where a preview view immediately reflects what a user is editing in a form.


[Here is a solid example](https://fburl.com/diffusion/mptmlybb) from the events team.

Relay Environment for local preview:

* The preview view should commit updates to a separate relay environment to avoid overwriting server data with local updates.


* Use `CometEnvironmentFactory.createLocalEnvironment` or `createFBLocalEnvironment` to create a unique local Relay environment.


Creating the environment:

```js
const {createLocalEnvironment} = require('CometRelayEnvironmentFactory');
const eventPreviewEnvironment = createLocalEnvironment();`
```

 Example: https://fburl.com/ins2k98p

Rendering the preview:

* The preview should be rendered using `useLazyLoadQuery_DEPRECATED` with `store-only` fetch-policy, so that it can read the data from the local environment without sending requests to the server.
* A query should be defined for the preview, so that Relay can use it to read and updates the data.
* Adding `@raw_response_type` onto the query generates the flow type for the payload for updating the data.

```js
const data = useLazyLoadQuery_DEPRECATED<EventCometFormPreviewRootQuery>(
  previewQuery,
  {
    eventID: EVENT_FORM_PREVIEW_ID,
    isPreview: true,
    scale: WebPixelRatio.get(),
  },
  {fetchPolicy: 'store-only'},
);
```

Example: https://fburl.com/1zki95ng

Updating the preview:

* Store form states in React state(reducer) or context, then construct the full payload data for the preview query defined above.
* Use `commitPayload` method on the local environment to update the data for the preview.
* Use `generateUniqueClientID` from `CometRelay` or `relay-runtime` to create unique ids for the id fields in the payload, so that all local updates won't cause conflict ids in relay store.
* Use `RawResponseType` in the generated query file to flow type the payload, this avoids missing fields or wrong fields.
* Use `createPayloadFor3DField(fragmentName, normalizationFileResource, componentFileResource, restOfTheResponse)` to provide the field with 3D dependencies. With `RawResponseType` flow will statically make sure the document name and the rest of the response is correct.

```js
const {
  createOperationDescriptor,
  generateUniqueClientID,
  createPayloadFor3DField,
} = require('relay-runtime'); // or require('CometRelay') for Comet
const JSResource = require('JSResource');

const EVENT_FORM_PREVIEW_ID = generateUniqueClientID();
const previewQueryOperation = createOperationDescriptor(EventCometFormPreviewRootQuery, {
  eventID: EVENT_FORM_PREVIEW_ID,
});
const rendererAST = JSResource('m#SomePreview_section$normalization.graphql');
const rendererComponent = JSResource('m#SomePreview.react');

const payload: PreviewRawResponseType = {
  id: EVENT_FORM_PREVIEW_ID,
  preview_section_renderer: createPayloadFor3DField(
    'PreviewFragmentName',
    rendererAST,
    rendererComponent,
    {
       __typename: 'SOME_TYPE',
       name: '<name>',
    }
  )
}

localEnvironment.commitPayload(previewQueryOperation, payload);

```

Example: https://fburl.com/r8tnzcxa

</FbInternalOnly>


<DocsRating />
